"""
Automatic 2D class selection tool.

MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from os import path, listdir
import h5py
from PIL import Image  # install it via pip install pillow
import numpy
import mrcfile

"""
The format of the .hf file is the following:
    ['MDF']['images']['i']['image']   where i is a number representing the i-th images
hence to get the images number 5:
    ['MDF']['images']['5']['image'][()]
"""


def checkfiles(path_to_files):
    """
    checks if the hdf files are in the correct path and returns True if all of them exists
    :param path_to_files:    list of paths
    :return:
    """
    if isinstance(path_to_files, (list, tuple)):
        for p in path_to_files:
            if not path.isfile(p):
                return False
    elif isinstance(path_to_files, str):
        return path.isfile(path_to_files)
    return True


def calc_2d_spectra(img):
    from scipy import fftpack
    import numpy as np

    F1 = fftpack.fft2(img)
    F2 = fftpack.fftshift(F1)
    psd2D = np.abs(F2) ** 2

    return psd2D


def getList_files(paths):
    """
    Returns the list of the valid hdf files in the given paths. It is called recursively
    :param paths: path or list of paths
    :return:
    """
    if isinstance(paths, str):
        paths = [paths]
    list_new_paths = list()
    iterate = False
    for p in paths:
        if path.isdir(p):
            iterate = True
            list_new_paths += [path.join(p, f) for f in listdir(p)]
        elif path.isfile(p):
            list_new_paths.append(p)
        else:
            print(
                "WARNING: The given path '"
                + str(p)
                + "' is not a folder or a file and it will be ignored"
            )
    if iterate is True:
        return getList_files(list_new_paths)
    return list_new_paths


def getList_relevant_files(path_to_files):
    """
    Check if the given files are hdf/mrcs with a valid format. Return The list of valid hdf
    :param path_to_files: list of all the files present in the folder (and subfolder)given from the user
    :return: list of valid hdf
    """

    return [
        path_to_file
        for path_to_file in path_to_files
        if path_to_file.endswith("mrcs")
        or path_to_file.endswith("mrc")
        or h5py.is_hdf5(path_to_file)
    ]


""" FUNCTION TO READ THE HDF"""


def get_key_list_images(path_to_file):
    """
    Returns the list of the keys representing the images in the hdf/mrcs file. It will be converted in list of integer
    :param path_to_file:
    :return:
    """
    print("Try to list images on", path_to_file)

    if path.isfile(path_to_file):
        filename_ext = path.basename(path_to_file).split(".")[1]
        if filename_ext == "hdf":
            try:
                with h5py.File(path_to_file, "r") as f:
                    data = [int(v) for v in list(f["MDF"]["images"])]
                return data
            except:
                print(
                    "WARNING in get_list_images: the file '"
                    + path_to_file
                    + " is not an HDF file with the following format:\n\t['MDF']['images']. It will be ignored"
                )
        elif filename_ext in ["mrcs", "mrc"]:
            try:
                if filename_ext == "mrcs":
                    with mrcfile.mmap(path_to_file, permissive=True, mode="r") as mrc:
                        data = list(range(mrc.header.nz))
                    return data
                else:
                    data = list(range(1))
                    return data
            except Exception as e:
                print(e)
                print(
                    "WARNING in get_list_images: the file '"
                    + path_to_file
                    + " is not an valid mrc file. It will be ignored"
                )


def getImages_fromList_key(file_index_tubles):
    """
    Returns the images in the hdf file (path_to_file) listed in (list_images)
    :param path_to_file:    path to hdf file
    :param list_images: list of keys of the DB. It is the output( or part of its) given from 'get_list_images'
    :return: Returns a list of numpy arrays
    """

    result_data = list()
    for path_to_file, list_images in file_index_tubles:
        data = list()
        if path.isfile(path_to_file):
            if path.basename(path_to_file).split(".")[1] == "hdf":
                try:
                    with h5py.File(path_to_file, driver="core") as f:
                        if isinstance(list_images, list) or isinstance(
                            list_images, tuple
                        ):
                            data = [
                                f["MDF"]["images"][str(i)]["image"][()]
                                for i in list_images
                            ]  # [()] is used instead of .value

                        elif isinstance(list_images, int):
                            data = f["MDF"]["images"][str(list_images)]["image"][()]
                        else:
                            print(
                                "\nERROR in getImages_fromList_key: invalid list_images, it should be a string or a list/tuple of strings:",
                                type(list_images),
                            )
                            print("you try to get the following images")
                            print(list_images)
                            exit()
                except Exception as e:
                    print(e)
                    print(
                        "\nERROR in getImages_fromList_key: the file '"
                        + path_to_file
                        + " is not an HDF file with the following format:\n\t['MDF']['images']['0']['image']"
                    )
                    print("you try to get the following images")
                    print(list_images)
                    print("there are " + str(len(f["MDF"]["images"])))
                    exit()
            elif path.basename(path_to_file).split(".")[1] in ["mrc", "mrcs"]:
                data = []
                with mrcfile.mmap(path_to_file, permissive=True, mode="r") as mrc:

                    if isinstance(list_images, int):
                        list_images = [list_images]

                    if isinstance(list_images, list) or isinstance(list_images, tuple):
                        if mrc.header.nz > 1:
                            data = [mrc.data[i] for i in list_images]
                        elif len(list_images) == 1:
                            data = mrc.data
        result_data.append(data)
    return result_data


def getImages_fromList_key_old(path_to_file, list_images):
    """
    Returns the images in the hdf file (path_to_file) listed in (list_images)
    :param path_to_file:    path to hdf file
    :param list_images: list of keys of the DB. It is the output( or part of its) given from 'get_list_images'
    :return: Returns a list of numpy arrays
    """
    data = list()
    if path.isfile(path_to_file):
        if path.basename(path_to_file).split(".")[1] == "hdf":
            try:
                with h5py.File(path_to_file, driver="core") as f:
                    if isinstance(list_images, list) or isinstance(list_images, tuple):
                        data = [
                            f["MDF"]["images"][str(i)]["image"][()] for i in list_images
                        ]  # [()] is used instead of .value
                    elif isinstance(list_images, int):
                        data = f["MDF"]["images"][str(list_images)]["image"][()]
                    else:
                        print(
                            "\nERROR in getImages_fromList_key: invalid list_images, it should be a string or a list/tuple of strings:",
                            type(list_images),
                        )
                        print("you try to get the following images")
                        print(list_images)
                        exit()
            except Exception as e:
                print(e)
                print(
                    "\nERROR in getImages_fromList_key: the file '"
                    + path_to_file
                    + " is not an HDF file with the following format:\n\t['MDF']['images']['0']['image']"
                )
                print("you try to get the following images")
                print(list_images)
                print("there are " + str(len(f["MDF"]["images"])))
                exit()
        elif path.basename(path_to_file).split(".")[1] in ["mrc", "mrcs"]:
            data = []
            with mrcfile.mmap(path_to_file, permissive=True, mode="r") as mrc:

                if isinstance(list_images, int):
                    list_images = [list_images]

                if isinstance(list_images, list) or isinstance(list_images, tuple):
                    if mrc.header.nz > 1:
                        data = [mrc.data[i] for i in list_images]
                    elif len(list_images) == 1:
                        data = [mrc.data]
    return data


""" FUNCTION TO MANIPULATE THE IMAGES"""


def resize_img(img, resize=(76, 76)):
    """
    Resize the given image into the given size
    :param img: as numpy array
    :param resize: resize size
    :return: return the resized img
    """
    im = Image.fromarray(img)
    return numpy.array(im.resize(resize, resample=Image.BILINEAR))


def normalize_img(img):
    """
    normalize the images in base of its mean and variance
    :param img:
    :return:
    """
    import numpy as np

    # img = img.astype(np.float64, copy=False)
    mean = numpy.mean(img)
    std = numpy.std(img)
    img = (img - mean) / std
    # img = img.astype(np.float32, copy=False)
    return img


def flip_img(img, t=None):
    """
    It flip the image in function of the given typ
    :param img:
    :param t: type of the flip
                1 --> flip over the row. Flipped array in up-down direction.(X)
                2 --> flip over the column Flipped array in right-left direction(Y)
                3 --> flip over the column and the row (X and Y)
                otherwise --> no flip
    :return:
    """
    if t == 1:
        return numpy.flipud(img)
    elif t == 2:
        return numpy.fliplr(img)
    elif t == 3:
        return numpy.flipud(numpy.fliplr(img))
    return img



"""
Automatic 2D class selection tool.

MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
import os
import h5py
import numpy as np
import mrcfile

# pylint: disable=C0330, C0301

def write_labeled_hdf(results, output_path, filename):
    """
    The code will generate 2 files with 'output_path' name and suffix '_good' where it'll storage the good results
                            and '_bad' to storage the bad results
    :param results: List of results tuples. The tubles have the format (image_path,index_in_hdf,label,confidence)
    :param output_path: Path where the results should be written.
    :return: None
    """

    def write_hdf(original_path, out_path, image_indices):
        """
        It copies the image from the original file to the result files.
        :param original_path: hdf starting file
        :param out_path: hdf results file
        :param image_indices: list of index of the images to copy
        :return: None
        """
        counter = 0
        with h5py.File(original_path, driver="core") as original:
            with h5py.File(out_path, "w") as f:
                group = f.create_group("MDF/images/")

                group.attrs["imageid_max"] = np.array([len(image_indices) - 1], dtype=np.int32)

                for original_index in image_indices:
                    subgroup = f.create_group("MDF/images/" + str(counter) + "/")
                    subgroup.create_dataset(
                        "image",
                        data=original["MDF"]["images"][str(original_index)]["image"][
                            ()
                        ],
                    )
                    for k, v in original["MDF"]["images"][
                        str(original_index)
                    ].attrs.items():
                        subgroup.attrs[k] = v
                    counter += 1

    def write_mrcs(original_path, out_path, l):
        """

        :param original_path:
        :param out_path:
        :param l:
        :return:
        """
        if l:
            with mrcfile.mmap(original_path, permissive=True, mode="r") as original_mrc:
                with mrcfile.new(out_path) as new_mrc:
                    new_mrc.set_data(original_mrc.data[l])
    def write_mrc(out_path, results):
        f = open(out_path, "a+")
        f.write(results[0]+"\n")


    # the given results belong at one starting file
    # Sort results
    results = sorted(results, key=lambda x: x[0])

    # Group them by filename
    image_paths = [result[0] for result in results]
    import itertools
    running_index = 0
    for index, (key, group) in enumerate(itertools.groupby(image_paths)):
        grp = list(group)
        good_index = []
        bad_index = []
        path_original = key
        for _, _ in enumerate(grp):

            if results[running_index][2]==1:
                good_index.append(results[running_index][1])
            else:
                bad_index.append(results[running_index][1])

            running_index = running_index+1
            #print("test", )
            #good_index = [result[1] for result in results[index] if result[2] == 1]
            #bad_index = [result[1] for result in results[index] if result[2] == 0]

        if os.path.basename(path_original).split(".")[1] == "hdf":
            write_hdf(
                path_original, os.path.join(output_path, filename + "_good.hdf"), good_index
            )
            write_hdf(
                path_original, os.path.join(output_path, filename + "_bad.hdf"), bad_index
            )
        elif os.path.basename(path_original).split(".")[1] == "mrcs":
            write_mrcs(
                path_original,
                os.path.join(output_path, filename + "_good.mrcs"),
                good_index,
            )
            write_mrcs(
                path_original, os.path.join(output_path, filename + "_bad.mrcs"), bad_index
            )
        else:
            if results[running_index - 1][2]==1:
                write_mrc(os.path.join(output_path,"good.txt"),results[running_index-1])
            else:
                write_mrc(os.path.join(output_path, "bad.txt"), results[running_index - 1])




"""
Automatic 2D class selection tool.

MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
# pylint: disable=C0330, C0301

import os
import multiprocessing
from keras.utils import Sequence
from keras.layers import (
    Conv2D,
    Input,
    MaxPooling2D,
    BatchNormalization,
    Dense,
    UpSampling2D,
    GlobalAveragePooling2D,
    Dropout,
)
from keras.layers.merge import concatenate
from keras.layers.advanced_activations import LeakyReLU
from keras.models import Model
from keras.optimizers import Adam
from keras import backend as K
from keras.callbacks import (
    ModelCheckpoint,
    EarlyStopping,
    ReduceLROnPlateau,
    TensorBoard,
)

# from swa.keras import SWA
import numpy as np
from Auto2DSelect.SGDRSchedular import SGDRScheduler
from .helper import (
    getImages_fromList_key,
    resize_img,
    normalize_img,
    get_key_list_images,
    getList_relevant_files,
    getList_files,
    calc_2d_spectra
)
from .augmentation import Augmentation


class BatchGenerator(Sequence):
    """
    This is a generator for batches of training data. It reads in a selection of images (depending
    on the batch size), augment the images and return them together with target values.
    """

    def __init__(
        self, labeled_data, name, batch_size, input_image_shape, is_grey=False
    ):
        self.labeled_data = labeled_data
        self.name = name
        self.batch_size = batch_size
        self.input_image_shape = input_image_shape
        self.do_augmentation = name == "train"
        self.augmenter = Augmentation(is_grey)
        self.write_psd = False
        self.do_psd = False

    def __len__(self):
        """
        :return: Return the total number of batches to cover the whole training set.
        """
        num_batches = int(np.ceil(len(self.labeled_data) / self.batch_size))
        return num_batches

    def __getitem__(self, idx):
        """
        :return: Returns the batch idx.
                first output is a numpy array of images
                second output is a list of integer, which represents the label of the images
        """
        if not self.labeled_data:
            print("ERROR: the labeled data set is empty!")
            exit(-1)

        idx = idx % self.__len__()
        start = 0 if idx == 0 else (self.batch_size * idx) - 1
        end = start + self.batch_size
        batch_tubles = self.labeled_data[start:end]

        # Find unique hdf files and read the image
        unique_class_files = {data_tuble[0] for data_tuble in batch_tubles}
        images = []
        psds = []
        labels = []
        for class_file_path in unique_class_files:
            indicis_for_class_file = [
                data_tuble[1]
                for data_tuble in batch_tubles
                if data_tuble[0] == class_file_path
            ]
            labels_for_class_file = [
                data_tuble[2]
                for data_tuble in batch_tubles
                if data_tuble[0] == class_file_path
            ]
            class_index_tubles = [(class_file_path, indi) for indi in indicis_for_class_file]
            images_from_file = getImages_fromList_key(class_index_tubles)

            images_from_file = [
                resize_img(img, (self.input_image_shape[0], self.input_image_shape[1]))
                for img in images_from_file
            ]  # 2. Downsize images to network input size

            images = images + images_from_file

            labels = labels + labels_for_class_file



        if self.do_augmentation is True:

            images = [
                self.augmenter.image_augmentation(img) for img in images
            ]  # 3. Do data augmentation (+ flip image randomly (X,Y,TH, NONE)) but only for training, not for validation



        images = [
            normalize_img(img) for img in images
        ]  # 4. Normalize images ( subtract mean, divide by standard deviation)
        if self.do_psd:
            psds = [calc_2d_spectra(img) for img in images]
            if self.write_psd:
                np.savetxt("psd.txt",psds[0])
            psds = [
                normalize_img(psd) for psd in psds
            ]
            if self.write_psd:
                from tifffile import imsave
                imsave('image.tif', images[0])
                #np.savetxt("image.txt",images[0])
                np.savetxt("psd_norm.txt",psds[0])
                self.write_psd=False

        images = np.array(images)
        if self.do_psd:
            images = np.stack((images, psds), axis=3)
        else:
            images = images[:, :, :, np.newaxis]

        # print(str(idx)+self.name)
        return images, labels

    def on_epoch_end(self):
        """
        This method shuffle the training data at the end of a epoch.
        :return: None
        """
        np.random.shuffle(self.labeled_data)


class Auto2DSelectNet:
    """
    Network class for cinderella
    """

    def __init__(self, batch_size, input_size, depth=1):
        """

        :param batch_size: Batch size for training / prediction
        :param input_size: Input image size
        """
        self.batch_size = batch_size
        self.input_size = input_size
        self.model = self.build_phosnet_model(depth)

    def build_phosnet_model(self,depth):
        """

        :param input_size: Image input size
        :return: A keras model
        """
        input_image = Input(shape=(self.input_size[0], self.input_size[1], depth))

        # name of first layer
        if depth == 1:
            name_first_layer = "conv_1_depth1"
        else:
            name_first_layer = "bablub"

        # Layer 1
        layer_out = Conv2D(
            32,
            (3, 3),
            strides=(1, 1),
            padding="same",
            name=name_first_layer,
            use_bias=False,
        )(input_image)
        layer_out = BatchNormalization(name="norm_1")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)
        layer_out = MaxPooling2D(pool_size=(2, 2))(layer_out)

        # Layer 2
        layer_out = Conv2D(
            64, (3, 3), strides=(1, 1), padding="same", name="conv_2", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_2")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)
        layer_out = MaxPooling2D(pool_size=(2, 2))(layer_out)

        # Layer 3
        layer_out = Conv2D(
            128, (3, 3), strides=(1, 1), padding="same", name="conv_3", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_3")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 4
        layer_out = Conv2D(
            64, (1, 1), strides=(1, 1), padding="same", name="conv_4", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_4")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 5
        layer_out = Conv2D(
            128, (3, 3), strides=(1, 1), padding="same", name="conv_5", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_5")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)
        layer_out = MaxPooling2D(pool_size=(2, 2))(layer_out)

        # Layer 6
        layer_out = Conv2D(
            256, (3, 3), strides=(1, 1), padding="same", name="conv_6", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_6")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 7
        layer_out = Conv2D(
            128, (1, 1), strides=(1, 1), padding="same", name="conv_7", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_7")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 8
        layer_out = Conv2D(
            256, (3, 3), strides=(1, 1), padding="same", name="conv_8", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_8")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)
        layer_out = MaxPooling2D(pool_size=(2, 2))(layer_out)

        # Layer 9
        layer_out = Conv2D(
            512, (3, 3), strides=(1, 1), padding="same", name="conv_9", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_9")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 10
        layer_out = Conv2D(
            256, (1, 1), strides=(1, 1), padding="same", name="conv_10", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_10")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 11
        layer_out = Conv2D(
            512, (3, 3), strides=(1, 1), padding="same", name="conv_11", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_11")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 12
        layer_out = Conv2D(
            256, (1, 1), strides=(1, 1), padding="same", name="conv_12", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_12")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 13
        layer_out = Conv2D(
            512, (3, 3), strides=(1, 1), padding="same", name="conv_13", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_13")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        skip_connection = layer_out

        layer_out = MaxPooling2D(pool_size=(2, 2))(layer_out)

        # Layer 14
        layer_out = Conv2D(
            1024, (3, 3), strides=(1, 1), padding="same", name="conv_14", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_14")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 15
        layer_out = Conv2D(
            512, (1, 1), strides=(1, 1), padding="same", name="conv_15", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_15")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 16
        layer_out = Conv2D(
            1024, (3, 3), strides=(1, 1), padding="same", name="conv_16", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_16")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 17
        layer_out = Conv2D(
            512, (1, 1), strides=(1, 1), padding="same", name="conv_17", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_17")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 18
        layer_out = Conv2D(
            1024, (3, 3), strides=(1, 1), padding="same", name="conv_18", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_18")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 19
        layer_out = Conv2D(
            1024, (3, 3), strides=(1, 1), padding="same", name="conv_19", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_19")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        # Layer 20
        layer_out = Conv2D(
            1024, (3, 3), strides=(1, 1), padding="same", name="conv_20", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_20")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        layer_out = UpSampling2D(size=(2, 2))(layer_out)

        skip_connection = Conv2D(
            256, (1, 1), strides=(1, 1), padding="same", name="conv_21_", use_bias=False
        )(skip_connection)
        skip_connection = BatchNormalization(name="norm_21_")(skip_connection)
        skip_connection = LeakyReLU(alpha=0.1)(skip_connection)
        layer_out = concatenate([skip_connection, layer_out])

        # Layer 21
        layer_out = Conv2D(
            1024, (3, 3), strides=(1, 1), padding="same", name="conv_22", use_bias=False
        )(layer_out)
        layer_out = BatchNormalization(name="norm_22")(layer_out)
        layer_out = LeakyReLU(alpha=0.1)(layer_out)

        feature_extractor = Model(input_image, layer_out)

        layer_out = GlobalAveragePooling2D()(feature_extractor(input_image))

        output = Dense(64, activation="relu", name="denseL1")(layer_out)
        # output = Dropout(0.2)(output)
        output = Dense(10, activation="relu", name="denseL2")(output)
        # output = Dropout(0.2)(output)
        output = Dense(1, activation="sigmoid", name="denseL3")(output)

        model = Model(input_image, output)
        model.summary()
        return model

    def load_weights(self, model_path):
        """
        Load the weights
        :param model_path: Path to .h5 model file
        :return: None
        """
        self.model.load_weights(model_path, by_name=True)

    # Define our custom loss function

    def train(
        self,
        good_path,
        bad_path,
        save_weights_name,
        learning_rate=10 ** -4,
        nb_epoch=50,
        nb_epoch_early=10,
        pretrained_weights=None,
        seed=10,
        train_val_thresh=0.8,
        max_valid_img_per_file=10,
        warmrestarts=True,
    ):
        """
        Train the network on 2D classes.

        :param good_path: Path to folder with good classes
        :param bad_path: Path to folder with bad classes
        :param save_weights_name: Filename of the modle file
        :param pretrained_weights: Filepath to pretrained weights
        :param seed: Seed for random number selection
        :return: None
        """
        np.random.seed(seed)

        if os.path.exists(pretrained_weights):
            print("Load pretrained weights", pretrained_weights)
            self.model.load_weights(pretrained_weights, by_name=True,skip_mismatch=True)

        """
        labeled_data = get_data_tubles(good_path, bad_path)
        train_valid_split = int(0.8 * len(labeled_data))

        np.random.shuffle(labeled_data)
        train_data = labeled_data[:train_valid_split]
        valid_data = labeled_data[train_valid_split:]
        """
        train_data, valid_data, weights = get_train_valid_tubles(
            good_path, bad_path, train_val_thresh, max_valid_img_per_file
        )

        train_generator = BatchGenerator(
            labeled_data=train_data,
            name="train",
            batch_size=self.batch_size,
            input_image_shape=self.input_size,
        )
        valid_generator = BatchGenerator(
            labeled_data=valid_data,
            name="valid",
            batch_size=self.batch_size,
            input_image_shape=self.input_size,
        )

        # Define callbacks
        all_callbacks = []
        checkpoint = ModelCheckpoint(
            save_weights_name,
            monitor="val_loss",
            verbose=1,
            save_best_only=True,
            save_weights_only=True,
            mode="min",
            period=1,
        )
        all_callbacks.append(checkpoint)

        early_stop = EarlyStopping(
            monitor="val_loss",
            min_delta=0.0005,
            patience=nb_epoch_early,
            mode="min",
            verbose=1,
        )
        all_callbacks.append(early_stop)
        try:
            os.makedirs(os.path.expanduser("logs/"))
        except:
            pass

        tb_counter = (
            len(
                [
                    log
                    for log in os.listdir(os.path.expanduser("logs/"))
                    if "cinderella" in log
                ]
            )
            + 1
        )
        tensorboard = TensorBoard(
            log_dir=os.path.expanduser("logs/") + "cinderella" + "_" + str(tb_counter),
            histogram_freq=0,
            write_graph=True,
            write_images=False,
        )
        all_callbacks.append(tensorboard)
        if not warmrestarts:
            reduce_lr_on_plateau = ReduceLROnPlateau(
                monitor="val_loss",
                factor=0.1,
                patience=int(nb_epoch_early * 0.6),
                verbose=1,
            )
            all_callbacks.append(reduce_lr_on_plateau)
        else:
            schedule = SGDRScheduler(
                min_lr=1e-7,
                max_lr=1e-4,
                steps_per_epoch=len(train_generator),
                lr_decay=0.9,
                cycle_length=5,
                mult_factor=1.5,
            )
            all_callbacks.append(schedule)
        # define swa callback
        """
        swa = SWA(start_epoch=5,
                  lr_schedule='cyclic',
                  swa_lr=learning_rate*0.1,
                  swa_lr2=learning_rate,
                  swa_freq=4,
                  verbose=1)
        """

        optimizer = Adam(
            lr=learning_rate, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0
        )

        self.model.compile(
            optimizer=optimizer, metrics=["accuracy"], loss="binary_crossentropy",
        )
        self.model.fit_generator(
            generator=train_generator,
            validation_data=valid_generator,
            workers=12,
            epochs=nb_epoch,
            callbacks=all_callbacks,
            max_queue_size=multiprocessing.cpu_count()//2,
            use_multiprocessing=False,
            class_weight=weights,
        )
        # average_filename = "average.h5"
        # self.model.save_weights(average_filename)

        import h5py

        with h5py.File(save_weights_name, mode="r+") as f:
            f["input_size"] = self.input_size

        # with h5py.File(average_filename, mode='r+') as f:
        #    f["input_size"] = self.input_size

        print("Meta data saved in model.")

    def predict(self, input_path, model_path, good_thresh=0.5):
        """
        Runs a trained model on a .hdf file containing 2D classes.

        :param input_path: Path to input hdf file
        :param model_path: Path to .h5 weights file.
        :param good_thresh: Threshold for selecting good classes
        :return: Return a list of tuples with the format (input_path, index_in_hdf, label, confidence)
        """
        self.load_weights(model_path)

        relevant_files = getList_relevant_files(getList_files(input_path))
        print("Relevant files:", len(relevant_files))
        files_to_classify = []
        for file_to_classify in relevant_files:
            files_to_classify += [(file_to_classify, index) for index in get_key_list_images(file_to_classify)]

        results = []
        from tqdm import tqdm
        for img_chunk in tqdm(list(chunks(files_to_classify, self.batch_size))):
            list_img = getImages_fromList_key(img_chunk)
            result = self.predict_np_list(list_img)
            results.append(result)

        result = np.concatenate(tuple(results))
        result_tuples = []
        for index, img_tuble in enumerate(files_to_classify):
            label = 0
            if result[index] > good_thresh:
                label = 1

            confidence = result[index]
            if result[index] <= good_thresh:
                confidence = 1 - confidence
            result_tuples.append((img_tuble[0], img_tuble[1], label, confidence))

        return result_tuples

    def predict_np_arr(self, images):
        """
        Run the prediction on a 3D with format numpy array [IMDEX_INDEX, IMAGE_WIDTH, IMAGE_HEIGHT].
        :param images: numpy array in the format [IMDEX_INDEX, IMAGE_WIDTH, IMAGE_HEIGHT]
        :return: 1D numpy array with probability of being a good class
        """
        list_img = [images[i] for i in range(images.shape[0])]
        return self.predict_np_list(self, list_img)

    def predict_np_list(self, list_img, do_psd=False):
        """
        Run the prediction on list of 2d numpy arrays.

        :param list_img: List of 2d numpy arrays (images)
        :return: 1D numpy array with probability of being a good class
        """
        list_img = [
            resize_img(img, (self.input_size[0], self.input_size[1]))
            for img in list_img
        ]  # 2. Downsize images to network input size
        list_img = [normalize_img(img) for img in list_img]

        arr_img = np.array(list_img)
        arr_img = np.expand_dims(arr_img, 3)
        pred_res = self.model.predict(arr_img)
        result = pred_res[:, 0]
        return result


def get_data_tubles(good_path, bad_path):
    """
    :param good_path: Path to the folder with good classes
    :param bad_path: Path to the folder with bad classes
    :return: List of tubles (HDF_PATH,CLASS_INDEX,LABEL) LABEL =1 good, 0 bad
    """
    list_bad = list()
    list_good = list()
    for good_p in getList_relevant_files(getList_files(good_path)):
        list_good += [(good_p, index, 1.0) for index in get_key_list_images(good_p)]
    for bad_p in getList_relevant_files(getList_files(bad_path)):
        list_bad += [(bad_p, index, 0.0) for index in get_key_list_images(bad_p)]
    return list_good + list_bad


def get_train_valid_tubles(good_path, bad_path, thresh=0.9, max_val_img_per_file=-1):
    list_bad_train = list()
    list_good_train = list()
    list_bad_valid = list()
    list_good_valid = list()

    for good_p in getList_relevant_files(getList_files(good_path)):
        good_tubles = [(good_p, index, 1.0) for index in get_key_list_images(good_p)]
        if len(good_tubles)>1:
            train_valid_split = int(thresh * len(good_tubles))
            if max_val_img_per_file > -1:
                train_valid_split = max(
                    train_valid_split, len(good_tubles) - max_val_img_per_file
                )
            np.random.shuffle(good_tubles)

            # Do the valid/train split for each file
            list_good_train += good_tubles[:train_valid_split]
            list_good_valid += good_tubles[train_valid_split:]
        else:
            if np.random.rand() > thresh:
                list_good_valid.extend(good_tubles)
            else:
                list_good_train.extend(good_tubles)

    for bad_p in getList_relevant_files(getList_files(bad_path)):
        bad_tubles = [(bad_p, index, 0.0) for index in get_key_list_images(bad_p)]
        if len(bad_tubles) > 1:
            train_valid_split = int(thresh * len(bad_tubles))
            if max_val_img_per_file > -1:
                train_valid_split = max(
                    train_valid_split, len(bad_tubles) - max_val_img_per_file
                )
            np.random.shuffle(bad_tubles)

            list_bad_train += bad_tubles[:train_valid_split]
            list_bad_valid += bad_tubles[train_valid_split:]
        else:
            if np.random.rand() > thresh:
                list_bad_valid.extend(bad_tubles)
            else:
                list_bad_train.extend(bad_tubles)
    print("CLASS 1", len(list_good_train), "CLASS 0", len(list_bad_train))

    list_train = list_good_train + list_bad_train
    list_valid = list_good_valid + list_bad_valid
    weight_good = 1 - float(len(list_good_train)) / len(list_train)
    weight_bad = 1 - float(len(list_bad_train)) / len(list_train)
    return list_train, list_valid, (weight_bad, weight_good)


def chunks(list_to_divide, number_of_chunks):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(list_to_divide), number_of_chunks):
        yield list_to_divide[i : i + number_of_chunks]


def focal_loss(y_true, y_pred):
    """Focal loss for multi-classification
    FL(p_t)=-alpha(1-p_t)^{gamma}ln(p_t)
    Notice: y_pred is probability after softmax
    gradient is d(Fl)/d(p_t) not d(Fl)/d(x) as described in paper
    d(Fl)/d(p_t) * [p_t(1-p_t)] = d(Fl)/d(x)
    Focal Loss for Dense Object Detection
    https://arxiv.org/abs/1708.02002

    Arguments:
        y_true {tensor} -- ground truth labels, shape of [batch_size, num_cls]
        y_pred {tensor} -- model's output, shape of [batch_size, num_cls]

    Keyword Arguments:
        gamma {float} -- (default: {2.0})
        alpha {float} -- (default: {4.0})

    Returns:
        [tensor] -- loss.
    """
    import tensorflow as tf

    gamma = 2.0
    alpha = 4.0
    epsilon = 1.0e-9
    y_true = tf.convert_to_tensor(y_true, tf.float32)
    y_pred = tf.convert_to_tensor(y_pred, tf.float32)

    model_out = tf.add(y_pred, epsilon)
    ce = tf.multiply(y_true, -tf.log(model_out))
    weight = tf.multiply(y_true, tf.pow(tf.subtract(1.0, model_out), gamma))
    fl = tf.multiply(alpha, tf.multiply(weight, ce))
    reduced_fl = tf.reduce_max(fl, axis=1)
    return tf.reduce_mean(reduced_fl)


def focal_loss(y_true, y_pred):
    import tensorflow as tf

    gamma = 2.0
    alpha = 4.0
    # epsilon = 1.e-9
    # y_pred = tf.add(y_pred, epsilon)
    pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
    pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))
    return -K.sum(alpha * K.pow(1.0 - pt_1, gamma) * K.log(pt_1)) - K.sum(
        alpha * K.pow(pt_0, gamma) * K.log(1.0 - pt_0)
    )



#! /usr/bin/env python
'''
Automatic 2D class selection tool.

MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''
import json
import multiprocessing
import argparse
import os
from . import results_writer
import h5py

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
try:
    os.environ["CUDA_VISIBLE_DEVICES"]
except KeyError:
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

ARGPARSER = argparse.ArgumentParser(
    description="Train auto 2D class selection",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)


ARGPARSER.add_argument("-i", "--input", required=True, help="Path folder to input HDF.")

ARGPARSER.add_argument("-o", "--output", required=True, help="Path to output folder.")

ARGPARSER.add_argument("-w", "--weights", required=True, help="Path network weights.")

ARGPARSER.add_argument(
    "-t",
    "--confidence_threshold",
    default=0.5,
    type=float,
    help="Classes with a confidence higher as that threshold are classified as good.",
)

ARGPARSER.add_argument("--gpu", default=-1, type=int, help="GPU to run on.")

ARGPARSER.add_argument("-b","--batch_size", default=32, type=int, help="Number of mini-batches during prediction.")


def _main_():
    args = ARGPARSER.parse_args()
    input_path = args.input
    weights_path = args.weights
    output_path = args.output
    threshold = args.confidence_threshold
    if os.path.exists(output_path):
        print("Output path already exists. Stop")
        exit(0)
    else:
        os.makedirs(output_path)

    if args.gpu != -1:
        str_gpu = str(args.gpu)
        os.environ["CUDA_VISIBLE_DEVICES"] = str_gpu

    with h5py.File(weights_path, mode="r") as f:
        try:
            import numpy as np
            input_size = tuple(f["input_size"])
        except KeyError:
            print("Did not found input size in your model file. Did you train JANNI with older version than 0.3? In that case, please retrain!")
            import sys
            sys.exit(0)

    batch_size = args.batch_size
    from .auto_2d_select import Auto2DSelectNet
    auto2dnet = Auto2DSelectNet(batch_size, input_size)
    result = auto2dnet.predict(input_path, weights_path, good_thresh=threshold)


    results_writer.write_labeled_hdf(
        result, output_path, os.path.basename(input_path).split(".")[0]
    )

    good = []
    bad = []
    for res in result:
        if res[2]:
            good.append(res[1])
        else:
            bad.append(res[1])
    good.sort()
    bad.sort()

    fraction_good = int(100*(len(good) / (len(good)+len(bad))))
    fraction_bad = int(100 * (len(bad) / (len(good) + len(bad))))
    print("\n Good classes: ", len(good), "/", len(good)+len(bad),"(",fraction_good,"% )",  "\n")
    print("\n Bad classes: ", len(bad), "/", len(good)+len(bad),"(",fraction_bad,"% )",  "\n")


if __name__ == "__main__":
    multiprocessing.set_start_method("spawn")
    _main_()



'''
Automatic 2D class selection tool.

MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import random
import numpy as np
from scipy import ndimage
from . import helper


class Augmentation:
    """
    Class for doing data augmentation
    """

    def __init__(self, is_grey=False):
        self.is_grey = is_grey

    def augmentation_pipeline(self, image, mandatory_augs, optional_augs, max_optional=6):

        # Run mandatory augmentations
        image = image.astype(np.float32, copy=False)
        for aug in mandatory_augs:
            image = aug(image)

        # Run optional augmentatiojs
        num_augs = np.random.randint(0, np.minimum(max_optional, len(optional_augs)))
        if num_augs > 0:
            selected_augs = random.sample(optional_augs, num_augs)
            for aug in selected_augs:
                image = aug(image)

        return image

    def image_augmentation(self, image):
        """
        Applies random selection of data augmentations
        :param image:  Input image
        :return: Augmented image
        """
        optional_augmentations = [
            self.additive_gauss_noise,
            self.add,
            self.contrast_normalization,
            self.multiply,
            # self.sharpen,
            self.dropout,
        ]
        r_blurr = np.random.rand()
        if r_blurr < 0.3:
            optional_augmentations.append(self.gauss_blur)
        elif r_blurr < 0.6:
            optional_augmentations.append(self.avg_blur)


        mandatory_augmentations = [self.flip,self.rotate]

        image = self.augmentation_pipeline(image, mandatory_augmentations, optional_augmentations)

        '''
        num_augs = np.random.randint(0, np.minimum(6, len(augmentations)))
        if num_augs > 0:
            r_blurr = np.random.rand()
            if r_blurr < 0.3:
                augmentations.append(self.gauss_blur)
            elif r_blurr < 0.6:
                augmentations.append(self.avg_blur)

            selected_augs = random.sample(augmentations, num_augs)
            image = image.astype(np.float32, copy=False)
            for sel_aug in selected_augs:
                image = sel_aug(image)
            #   print "Mean after", sel_aug, " sum: ", np.mean(image)
            if self.is_grey:
                min_img = np.min(image)
                max_img = np.max(image)
                image = ((image - min_img) / (max_img - min_img)) * 255
                #    image = np.clip(image, 0, 255)
                image = image.astype(np.uint8, copy=False)

        image = self.flip_rotation_augmentation(image)
        '''

        return image


    def flip(self,image):
        # Random flip
        image = helper.flip_img(image, random.randint(0, 3))
        return image

    def rotate(self,image):
        # Random rotation
        rand_rotation = np.random.randint(4)
        image = np.rot90(image, k=rand_rotation)
        return image

    def gauss_blur(self, image, sigma_range=(0, 3)):
        """
        Applys gaussian blurring with random sigma
        :param image: Input image
        :param sigma_range:  Range for random sigma
        :return: Blurred image
        """
        rand_sigma = sigma_range[0] + np.random.rand() * (
            sigma_range[1] - sigma_range[0]
        )
        result = ndimage.gaussian_filter(image, sigma=rand_sigma, mode="nearest")

        if not np.issubdtype(image.dtype, np.float32):
            result = result.astype(np.float32, copy=False)
        return result

    def avg_blur(self, image, kernel_size=(2, 7)):
        """
        Applys average blurring with random kernel size
        :param image: Input image (numpy array)
        :param kernel_size: Range for random kernel size
        :return: Blurred image
        """
        rang_kernel_size = np.random.randint(kernel_size[0], kernel_size[1])
        image = ndimage.uniform_filter(image, size=rang_kernel_size, mode="nearest")
        return image

    def additive_gauss_noise(self, image, max_sigma_range_factor=0.05):
        """
        Add random gaussian noise to image
        :param image: Input image
        :param max_sigma_range_factor: Range for max_sigma_range. The standard deviation of the noise is
        choosen randomly depending on the standard deviation of the image.
        The choosen standard deviation for noise is between: 0 and  max_sigma_factor*6*np.std(image)
        :return:
        """

        width = 2 * 3 * np.std(image)
        max_sigma = width * max_sigma_range_factor
        rand_sigma = np.random.rand() * max_sigma

        noise = np.random.randn(image.shape[0], image.shape[1])
        np.multiply(noise, rand_sigma, out=noise)
        np.add(image, noise, out=image)

        if not np.issubdtype(image.dtype, np.float32):
            image = image.astype(np.float32, copy=False)

        return image

    def contrast_normalization(self, image, alpha_range=(0.5, 2.0)):
        """
        Spread or squeeze the pixel values.
        :param image: Input image
        :param alpha_range: Range for alpha. Alpha controlls the normalization.
        :return: Modified image
        """
        rand_multiplyer = alpha_range[0] + np.random.rand() * (
            alpha_range[1] - alpha_range[0]
        )
        middle = np.median(image)
        np.subtract(image, middle, out=image)
        np.multiply(rand_multiplyer, image, out=image)
        np.add(middle, image, out=image)

        return image

    def dropout(self, image, ratio=(0.01, 0.1)):
        """
        Set a random selection of pixels to the mean of the image
        :param image: Input image
        :param ratio: Range for random ratio
        :return: Modified image
        """
        if isinstance(ratio, float):
            rand_ratio = ratio
        else:
            rand_ratio = ratio[0] + np.random.rand() * (ratio[1] - ratio[0])
        mean_val = np.mean(image)
        drop = np.random.binomial(
            n=1, p=1 - rand_ratio, size=(image.shape[0], image.shape[1])
        )
        image[drop == 0] = mean_val

        return image

    def add(self, image, scale=0.05):
        """
        Adds a random constant to the image
        :param image: Input image
        :param scale: Scale for random constant. The random constant
        will be between 0 and scale*6*std(image)
        :return: Modified image
        """
        width = 2 * 3 * np.std(image)
        width_rand = scale * width
        rand_constant = (np.random.rand() * width_rand) - width_rand / 2
        np.add(image, rand_constant, out=image)

        return image

    def multiply(self, image, mult_range=(0.5, 1.5)):
        """
        Multipy the input image by a random float.
        :param image: Input image
        :param mult_range: Range for random multiplier
        :return: multiplied image
        """

        rand_multiplyer = mult_range[0] + np.random.rand() * (
            mult_range[1] - mult_range[0]
        )
        np.multiply(image, rand_multiplyer, out=image)
        return image
